﻿'   EXPCO2
'
'
'   DESCRIPTION
'
'   This program came about in order to see if an exponential curve would fit
'   the atmospheric CO2 levels. I had no idea if it would, or not, but felt
'   the question was interesting. Rates based upon concentrations abound in
'   nature, so exponentials are found almost everywhere in nature, as well.
'
'   The question of why a simple, single exponential curve might fit the
'   atmospheric concentrations of CO2 is a different one. On first blush, I'd
'   imagine that there are many complex interactions, each with their own
'   individual timing relationships, and of course there is life itself which
'   can add unique and complicating aspects. But there is a massive driving
'   force with anthropogenic CO2 and this is, of course, humans themselves and
'   their population on the planet. Population drives almost everything and
'   instantaneous rate remains proportional to concentration (population) even
'   if the proportional factor itself varies over time. So it does suggest the
'   idea that anthropogenic CO2, and therefore as some long consequential
'   chain that follows, may still have some relationship to an exponential
'   curve; even with so many possible confounding factors.
'
'   The question is interesting in another regard. There are three factors to
'   solve for a simple exponential model. Offset, gain, and the tau. In this
'   context, the idea of 'offset' relates back to the pre-industrial levels of
'   atmospheric CO2, which is a background level prior to industrial use of
'   fossil fuels by humans. If such a 'natural equilibrium' state exists, or
'   as Dr. Keeling called it, a 'characteristic background concentration,'
'   and if the process matches closely to an exponential model, then we'd
'   predict (expect) that a better offset value will correspond to estimates
'   of this pre-industrial equilibrium state.
'
'   This pre-industrial value is often estimated to be around 280 ppm because
'   it is close to the average of CO2 measured and dated using high time
'   resolution between the years 1000 And 1800 in an ice core from Law Dome,
'   Antarctica. [Etheridge et al., 1996]. So if we should find an offset that
'   is close to this value then that would tend to confirm our hypothesis. If
'   not, that would tend to dis-confirm it.
'
'
'   SOURCE DATA
'
'   For the most part, this is a very simple program and I wanted to read up a
'   data file (ASCII) from some site. So the first thing I needed to do was to
'   identify the specific file. Mauna Loa in Hawaii has the longest record and
'   often is the "go to" site. So I decided to use that one. Today, they take
'   hourly flask samples. But I couldn't find that data set, readily. But I
'   was able to find monthly averages and that's fine. Monthly data will show
'   the annual cycles quite nicely and won't interfere much with a quality
'   fitting algorithm. The file I finally used comes from this site:
'
'       ftp://aftp.cmdl.noaa.gov/products/trends/co2/co2_mm_mlo.txt
'
'
'   DESIGN
'
'   The program uses a file-open dialog box to ask the user for the above file
'   to use. That should be placed somewhere convenient. The program then reads
'   up the data and uses the 'interpolated' column of data (to avoid -1 values
'   that may appear in the 'average' column.) No dates are kept associated to
'   the data values, as it's not important. With the data loaded up, the code
'   needs to compute the exponential curve and display the raw data along with
'   this new curve. The complexity in this code arrives from three problems:
'
'       (1)     Computing the best fitted exponential 
'       (2)     Computing nice chart boundaries and ticks
'       (3)     Generating a log-lin chart with appropriate grid
'                   lines and labels
'
'   Most of the above is technical programming stuff, so I'll leave others to
'   look over the code that does it. But the one thing that really matters is
'   the exponential fitting process. So I'll focus on that.
'
'   Probably one of the more important issues is the exact fitting algorithm
'   for fitting an exponential. Very commonly, the data is first converted by
'   taking the logarithm of the data and then using a very simple, commonly
'   found algorithm for fitting straight lines called "least squares fitting.'
'   However, there is a small problem. Measurements taken of raw data in the
'   exponential domain usually include 'noise.' This noise will usually be of
'   Gaussian shape and will tend to have constant characteristics independent
'   of the signal portion of the measurement. For example, a voltmeter will
'   have a certain noise level that is independent of measuring a 1V or 6V
'   source. When the data is then mapped into the log-domain, this noise is
'   no longer uniform. Measurements very close to the noise floor will be
'   treated the same as measurements far away, by naive least square fits.
'   So I've provided two different algorithms here. One that provides the
'   naive treatment (uniform treatment) and one which provides a weighted
'   treatment that gives high signal values more weight in establishing the
'   curve fit and that gives low signal values closer to the floor less
'   weight. Which is used is a matter of choice, as both are provided. There
'   is a group box with radio buttons on it to allow the selection of uniform
'   or weighted treatment.
'
'   The background equilibrium state is provided in a text box where the pre-
'   industrial estimate can be provided. (The fitting code doesn't attempt to
'   reconstruct this on its own, being based on a linear least square fit.)
'   Different values here will yield quite different exponentials. Play with
'   this value and see the effects. Using a blank (or zero) value will cause
'   the resulting exponential to be displayed as a perfectly straight line
'   on the log-linear chart. Adding a non-zero equilibrium state value will
'   cause the fitted line to become non-linear on the chart. But it may fit
'   the data better. Play with it.
'
Imports System.IO
Imports System.Math
Imports System.Windows.Forms.DataVisualization.Charting

Public Class ExpCO2

    Private Const DETRENDBASE As Double = 350.0
    Private Const FILTERBASE As Double = 320.0

    Private dataset As Double() = Nothing
    Private datasetmin As Double, datasetmax As Double
    Private Shared ticks As Double() = New Double() {10, 20, 25, 30, 40, 50, 60, 70, 75, 80, 90, 100}

    Private Class AnnualFilter
        Private Shared filtertaps As Double() = New Double() {0.0026144372353078377, 0.00099477395904088448,
        0.0005009099623101271, -0.00045994367198163661, -0.0016920488303780999, -0.0028609343907711189,
       -0.0035592999495704771, -0.0034284949653726822, -0.0022935620263246042, -0.00027063747314539694,
        0.0022063529336049433, 0.0044581309695086446, 0.0057434292659487634, 0.0054915720408955744,
        0.0035168025431912184, 0.0001328087372984035, -0.0038776881044137648, -0.0073819076104813782,
       -0.0092402554523354219, -0.0086347463865987985, -0.0053866131857676766, -0.000091646379243009849,
        0.0059524258411883576, 0.011059730216280218, 0.013613524773627919, 0.012563145830053572,
        0.0077949216326646, 0.00028410545162081973, -0.0080842274729518229, -0.014977001547268149,
       -0.018267536892541427, -0.01668049459901182, -0.010232455769045971, -0.00035602167508108813,
        0.010411463324240283, 0.019075387039220564, 0.023020734383792728, 0.020794647656923133,
        0.012613444421995483, 0.00040311579536058165, -0.012643856582618294, -0.022914305595918179,
       -0.027371624293450521, -0.024469115123303832, -0.014661493624545923, -0.00037788770710964111,
        0.014589008932018116, 0.026104439609955364, 0.030845881596298568, 0.027272622485921688,
        0.01611982766944221, 0.0002552088327995479, -0.016062013546756206, -0.02834188347107475,
       -0.033123600353438233, -0.028964839559665893, -0.016879979989060781, -0.000083370393965438244,
        0.016869701028909712, 0.029347979806253154, 0.033926781057316245, 0.029347979806253154,
        0.016869701028909712, -0.000083370393965438244, -0.016879979989060781, -0.028964839559665893,
       -0.033123600353438233, -0.02834188347107475, -0.016062013546756206, 0.0002552088327995479,
        0.01611982766944221, 0.027272622485921688, 0.030845881596298568, 0.026104439609955364,
        0.014589008932018116, -0.00037788770710964111, -0.014661493624545923, -0.024469115123303832,
       -0.027371624293450521, -0.022914305595918179, -0.012643856582618294, 0.00040311579536058165,
        0.012613444421995483, 0.020794647656923133, 0.023020734383792728, 0.019075387039220564,
        0.010411463324240283, -0.00035602167508108813, -0.010232455769045971, -0.01668049459901182,
       -0.018267536892541427, -0.014977001547268149, -0.0080842274729518229, 0.00028410545162081973,
        0.0077949216326646, 0.012563145830053572, 0.013613524773627919, 0.011059730216280218,
        0.0059524258411883576, -0.000091646379243009849, -0.0053866131857676766, -0.0086347463865987985,
       -0.0092402554523354219, -0.0073819076104813782, -0.0038776881044137648, 0.0001328087372984035,
        0.0035168025431912184, 0.0054915720408955744, 0.0057434292659487634, 0.0044581309695086446,
        0.0022063529336049433, -0.00027063747314539694, -0.0022935620263246042, -0.0034284949653726822,
       -0.0035592999495704771, -0.0028609343907711189, -0.0016920488303780999, -0.00045994367198163661,
        0.0005009099623101271, 0.00099477395904088448, 0.0026144372353078377}
        Private index As Integer
        Private history As Double()
        Public Sub init()
            index = 0
            history = New Double(filtertaps.Count - 1) {}
            For i As Integer = 0 To filtertaps.Count - 1
                history(i) = 0.0
            Next
        End Sub
        Public Sub putv(ByVal v As Double)
            history(index) = v
            index += 1
            If index = filtertaps.Count Then index = 0
        End Sub
        Public Function getv() As Double
            Dim acc As Double = 0.0
            Dim idx = index
            For i As Integer = 0 To filtertaps.Count - 1
                idx -= 1
                If idx < 0 Then idx = filtertaps.Count - 1
                acc += history(idx) * filtertaps(i)
            Next
            Return acc
        End Function
    End Class

    Private Sub ExpCO2_Load(sender As Object, e As EventArgs) Handles Me.Load
        FileNameTxt.Text = ""
        EntryCountTxt.Text = ""
        TauTxt.Text = ""
        GainTxt.Text = ""
        PropRadio.Checked = True
        ComputeBtn.Enabled = False
    End Sub

    Private Async Sub FileOpenBtn_Click(sender As Object, e As EventArgs) Handles FileOpenBtn.Click
        Dim opendlg As New OpenFileDialog
        opendlg.DefaultExt = "txt"
        opendlg.Filter = "txt files (*.txt)|*.txt|All files (*.*)|*.*"
        opendlg.FilterIndex = 1
        If opendlg.ShowDialog() = DialogResult.OK Then
            FileNameTxt.Text = ""
            Try
                dataset = Await ReadCO2DataAsync(opendlg.FileName)
                If dataset.Count > 0 Then
                    Dim vmin As Double = dataset(0), vmax As Double = dataset(0)
                    For i As Integer = 1 To dataset.Count - 1
                        If dataset(i) < vmin Then
                            vmin = dataset(i)
                        ElseIf dataset(i) > vmax Then
                            vmax = dataset(i)
                        End If
                    Next
                    datasetmin = vmin
                    datasetmax = vmax
                    ComputeBtn.Enabled = True
                Else
                    ComputeBtn.Enabled = False
                End If
                FileNameTxt.Text = opendlg.SafeFileName
            Catch ex As Exception
                MessageBox.Show("Cannot read file. Original error: " & ex.Message)
                dataset = Nothing
                ComputeBtn.Enabled = False
            Finally
            End Try
            EntryCountTxt.Text = IIf(dataset Is Nothing OrElse dataset.Count < 1, "<empty>", dataset.Count.ToString())
        End If
    End Sub

    Private Sub ComputeBtn_Click(sender As Object, e As EventArgs) Handles ComputeBtn.Click
        If dataset Is Nothing Then Exit Sub
        Dim offset As Double = Val(OffsetBox.Text)
        Dim tdata(dataset.Count - 1) As Double
        For i As Integer = 0 To tdata.Count - 1
            tdata(i) = dataset(i) - offset
        Next
        Dim c1 As Double = 0.0, c2 As Double = 0.0
        Dim status As Boolean
        If PropRadio.Checked Then
            status = ComputeTauWeighted(tdata, c1, c2)
        Else
            status = ComputeTauUniform(tdata, c1, c2)
        End If
        If status Then
            Dim gain As Double = Math.Exp(c1)
            Dim tau As Double = 1.0 / c2
            Dim cset(dataset.Count - 1) As Double
            For i As Integer = 0 To dataset.Count - 1
                cset(i) = offset + gain * Math.Exp(i / tau)
            Next
            ' DoStuff(cset)
            Dim mdata(cset.Count - 1) As Double
            Dim fdata(cset.Count - 1) As Double
            For i As Integer = 0 To cset.Count - 1
                mdata(i) = DETRENDBASE + dataset(i) - cset(i)
                fdata(i) = dataset(i) - cset(i)
            Next
            Dim f As AnnualFilter = New AnnualFilter
            f.init()
            For i As Integer = 0 To 10
                For j As Integer = 0 To 11
                    f.putv(fdata(j))
                Next
            Next
            For i As Integer = 0 To cset.Count - 1
                f.putv(fdata(i))
                fdata(i) = f.getv()
            Next
            Dim ppdata As List(Of Double) = New List(Of Double)
            Dim pp As Double = -1000.0, pm As Double = -1000.0, skip As Boolean = True
            For i As Integer = 1 To cset.Count - 2
                If fdata(i - 1) < fdata(i) AndAlso fdata(i + 1) < fdata(i) Then
                    If pm >= -500.0 Then
                        If skip Then
                            skip = False
                        Else
                            ppdata.Add(fdata(i) - pm)
                        End If
                    End If
                    pp = fdata(i)
                ElseIf fdata(i - 1) > fdata(i) AndAlso fdata(i + 1) > fdata(i) Then
                    If pp >= -500.0 Then
                        If skip Then
                            skip = False
                        Else
                            ppdata.Add(pp - fdata(i))
                        End If
                    End If
                    pm = fdata(i)
                End If
            Next
            For i As Integer = 0 To cset.Count - 1
                fdata(i) += FILTERBASE
            Next
            TauTxt.Text = tau.ToString()
            GainTxt.Text = gain.ToString()
            LineChart.Series(0).Points().DataBindY(dataset)
            LineChart.Series(1).Points().DataBindY(mdata)
            LineChart.Series(2).Points().DataBindY(cset)
            LineChart.Series(3).Points().DataBindY(ppdata.ToArray())

            Dim mintick As Double, maxtick As Double, steptick As Double
            NiceMinMaxStep(datasetmin, datasetmax, mintick, maxtick, steptick)

            LineChart.ChartAreas(0).AxisY.IsLogarithmic = True
            LineChart.ChartAreas(0).AxisY.LogarithmBase = 10
            LineChart.ChartAreas(0).AxisY.Minimum = mintick
            LineChart.ChartAreas(0).AxisY.Maximum = maxtick
            LineChart.ChartAreas(0).AxisY2.IsLabelAutoFit = False
            LineChart.ChartAreas(0).AxisY2.Minimum = 4.0
            LineChart.ChartAreas(0).AxisY2.Maximum = 8.0
            LineChart.ChartAreas(0).AxisY2.Interval = 1.0

            Dim xi As Double = Math.Log10(maxtick / mintick)
            Dim xj As Double = Math.Log10((mintick + steptick) / mintick)

            LineChart.ChartAreas(0).AxisY.IsLabelAutoFit = False
            LineChart.ChartAreas(0).AxisY.MajorGrid.Enabled = False
            For v As Double = mintick To maxtick Step steptick
                Dim cl As CustomLabel = New CustomLabel
                If v = mintick Then
                    cl.FromPosition = Math.Log10(v - 5)
                    cl.ToPosition = Math.Log10(v + 7)
                Else
                    cl.FromPosition = Math.Log10(v - 7)
                    cl.ToPosition = Math.Log10(v + 7)
                End If
                cl.Text = v.ToString()
                LineChart.ChartAreas(0).AxisY.CustomLabels.Add(cl)
            Next
            For v As Double = mintick + steptick To maxtick Step steptick
                Dim sl As StripLine = New StripLine
                sl.IntervalOffset = Math.Log10(v)
                sl.StripWidth = 0.0001
                sl.Interval = 0
                sl.BackColor = Color.Black
                LineChart.ChartAreas(0).AxisY.StripLines.Add(sl)
            Next
            LineChart.Titles(1).Text = "co2(t) = " & offset.ToString() & " + " & gain.ToString() & " " & ChrW(&HD7) & " e^( t / " & tau.ToString() & " )"
        Else
            TauTxt.Text = "Error"
            GainTxt.Text = "Error"
        End If
    End Sub

    Private Function ComputeTauUniform(data As Double(), ByRef c1 As Double, ByRef c2 As Double) As Boolean
        If data IsNot Nothing AndAlso data.Count > 0 Then
            Dim x As Double = 0.0, y As Double = 0.0, xy As Double = 0.0, xx As Double = 0.0
            For i As Integer = 0 To data.Count - 1
                Dim ln As Double = Math.Log(data(i))
                x += i
                y += ln
                xx += i * i
                xy += i * ln
            Next
            Dim num As Double = data.Count * xy - x * y
            Dim den As Double = data.Count * xx - x * x
            If den <> 0 Then
                c2 = num / den
                c1 = (y - c2 * x) / data.Count
                Return True
            End If
        End If
        Return False
    End Function

    Private Function ComputeTauWeighted(data As Double(), ByRef c1 As Double, ByRef c2 As Double) As Boolean
        If data IsNot Nothing AndAlso data.Count > 0 Then
            Dim x As Double = 0.0, y As Double = 0.0, xy As Double = 0.0, xx As Double = 0.0, n As Double = 0.0
            For i As Integer = 0 To data.Count - 1
                Dim z = data(i)
                Dim ln As Double = Math.Log(z)
                Dim w As Double = z * z
                n += w
                x += w * i
                y += w * ln
                xx += w * i * i
                xy += w * i * ln
            Next
            Dim num As Double = n * xy - x * y
            Dim den As Double = n * xx - x * x
            If den <> 0 Then
                c2 = num / den
                c1 = (y - c2 * x) / n
                Return True
            End If
        End If
        Return False
    End Function

    Private Async Function ReadCO2DataAsync(ByVal fname As String) As Task(Of Double())
        Dim results As List(Of Double) = New List(Of Double)
        Using sr As StreamReader = New StreamReader(fname)
            Dim curline As String = Await sr.ReadLineAsync()
            Do While curline IsNot Nothing
                If curline <> "" AndAlso Mid(curline, 1, 1) <> "#" Then
                    Dim words As String() = curline.Split(New Char() {" "}, StringSplitOptions.RemoveEmptyEntries)
                    If words.Count > 6 Then
                        Try
                            For i As Integer = 0 To words.Count - 1
                                Dim q As Double = Val(words(i))
                            Next
                            Dim v As Double = Val(words(4))
                            results.Add(v)
                        Catch ex As Exception
                            Return New Double() {}
                        Finally
                        End Try
                    Else
                        Return New Double() {}
                    End If
                End If
                curline = Await sr.ReadLineAsync()
            Loop
        End Using
        Return results.ToArray()
    End Function

    Private Function TickFind(ByVal v As Double) As Integer
        For i As Integer = 0 To ticks.Count - 1
            If ticks(i) > v AndAlso Math.Abs(ticks(i) - v) > 0.00000000001 Then Return i - 1
        Next
        Return -1
    End Function

    Private Sub NiceMinMaxStep(ByVal mn As Double, ByVal mx As Double, ByRef minv As Double, ByRef maxv As Double, ByRef stepv As Double)
        Dim pwrmn As Integer = CInt(Math.Floor(Math.Log10(mn))) - 1
        Dim imn As Integer = TickFind(mn / 10 ^ pwrmn)
        Dim pwrmx As Integer = CInt(Math.Floor(Math.Log10(mx))) - 1
        Dim imx As Integer = TickFind(mx / 10 ^ pwrmx)
        minv = ticks(imn) * 10.0 ^ pwrmn
        maxv = ticks(imx) * 10.0 ^ pwrmx
        Dim m As Double
        If imx = ticks.Count - 1 Then
            m = ticks(0) * 10.0 ^ (pwrmx + 1)
        Else
            m = ticks(imx + 1) * 10.0 ^ pwrmx
        End If
        Dim i As Integer = 0
        Do While i < 1 AndAlso (mx - mn) / (m - minv) < 0.8
            imn = TickFind(mn - minv)
            pwrmn -= 1
            If imn >= 0 Then
                minv += ticks(imn) * 10.0 ^ pwrmn
            End If
            imx = TickFind(mx - maxv)
            pwrmx -= 1
            m = maxv + ticks(imx + 1) * 10.0 ^ pwrmx
            If imx >= 0 Then
                maxv += ticks(imx) * 10.0 ^ pwrmx
            End If
            i += 1
        Loop
        maxv = m
        Dim stp As Double = (maxv - minv) / 6.0
        Dim pwr As Double = CInt(Math.Floor(Math.Log10(stp))) - 1
        Dim idx As Integer = TickFind(stp / 10 ^ pwr)
        stepv = ticks(idx) * 10.0 ^ pwr
    End Sub

    'Private Sub DoStuff(ByRef c As Double())
    '    Dim mdata(c.Count - 1) As Double
    '    For i As Integer = 0 To c.Count - 1
    '        mdata(i) = dataset(i) - c(i)
    '    Next
    '    Dim phase0 As Double = 0.0, phase90 As Double = 0.0, pcount As Integer = 0
    '    Dim plist As List(Of Double) = New List(Of Double)
    '    For i As Integer = dataset.Count - 1 To 11 Step -12
    '        pcount += 1
    '        Dim p0 As Double = 0.0, p90 As Double = 0.0
    '        For j As Integer = 11 To 0 Step -1
    '            p0 += mdata(i - j) * Math.Cos(Math.PI * (j - 11.0) / 6.0)
    '            p90 += mdata(i - j) * Math.Sin(Math.PI * (j - 11.0) / 6.0)
    '        Next
    '        plist.Add(Math.Sqrt(p0 * p0 + p90 * p90) / 3.0)
    '        ' plist.Add(Math.Sqrt(p0 * p0 + p90 * p90))
    '        phase0 += p0
    '        phase90 += p90
    '    Next
    '    Dim quadrature As Double = Math.Sqrt(phase0 * phase0 + phase90 * phase90) / pcount / 3.0
    '    Dim q0 As Double = 0.0, q90 As Double = 0.0, a As Double = 7.0
    '    For i As Integer = 0 To 11
    '        q0 += a * Math.Sin(Math.PI * i / 6.0) * Math.Cos(Math.PI * i / 6.0)
    '        q90 += a * Math.Sin(Math.PI * i / 6.0) * Math.Sin(Math.PI * i / 6.0)
    '    Next
    '    Dim quad As Double = Math.Sqrt(q0 * q0 + q90 * q90) / 3.0
    '    Dim myvalue As Double = 1.0
    'End Sub

End Class
